# -*- coding: utf-8 -*-

module ASN1
  class OutOfBytesError < RuntimeError
    def initialize(bytes)
      @bytes = bytes
    end

    def to_s
      "Out of bytes to decode: #{@bytes}"
    end
    attr_reader :bytes
  end

  def self.encode(d)
    Buffer.new.add(d).buffer
  end
  def self.encode_seq(ds)
    out = Buffer.new
    ds.each do |d|
      out.add(d)
    end
    out.buffer
  end
  def self.decode(s)
    Buffer.new(s).decode_item_final
  end
  def self.decode_with_rest(s)
    Buffer.new(s).decode_item_with_rest
  end
  def self.decode_seq(s)
    Buffer.new(s).decode_items
  end

  class Buffer

  attr_reader :buffer
  def initialize(s = String.new)
    @buffer = s
    @pos = 0
  end

  def atleast(n)
    left = @buffer.bytesize - @pos
    raise OutOfBytesError.new(n - left) if n > left
  end

  def take(n)
    opos = @pos
    @pos += n
    raise OutOfBytesError.new(@pos - @buffer.bytesize) if @pos > @buffer.bytesize
    @buffer[opos, n]
  end

  def decode_sdnv
    res = 0
    while (b = take(1).ord
           res = (res << 7) + (b & 0x7f)
           (b & 0x80) != 0)
      ;
    end
    res
  end

  def self.oid_print(s)
    bu = Buffer.new(s)
    a = []
    while !bu.done
      a << bu.decode_sdnv
    end
    ia = a[0]
    ia1 = [a[0]/40, 2].min
    a[0..0] = [ia1, ia - 40 * ia1]
    a.join(".")
  end

  def self.be_decode(s)
    ret = 0
    s.each_byte do |b|
      ret = (ret << 8) + b
    end
    ret
  end
  define_method :be_decode, &method(:be_decode)

  READABLE = Hash.new {|h, k| [k]}
  READABLE[[0x30, 0x10]] = :seq
  READABLE[[0x31, 0x11]] = :set
  READABLE[[1, 1]] = :bool, ->(s) {s != "\0"}
  READABLE[[2, 2]] = :int, ->(s) {be_decode(s)} # XXX: negative!
  READABLE[[3, 3]] = :bits
  READABLE[[4, 4]] = :b
  READABLE[[5, 5]] = :null, ->(s) {:null}
  READABLE[[6, 6]] = :oid, ->(s) {oid_print(s)}
  READABLE[[12, 12]] = :u
  READABLE[[19, 19]] = :s
  READABLE[[22, 22]] = :ia5
  READABLE[[23, 23]] = :t       # 2-digit year utc time
  READABLE[[24, 24]] = :gt       # generalized time
  READABLE[[30, 30]] = :ucs2, ->(s) { {ucs2: s.force_encoding(Encoding::UTF_16BE).encode(Encoding::UTF_8)} }       # BMPString
  (0..15).each do |i|
    READABLE[[0xA0 + i, i]] = :"exp#{i}"
  end

  def decode_item
    ib = take(1).ord
    cons = ib & 0x20
    tag = decode_sdnv if (tag = ib & 0x1f) == 0x1f
    case len = take(1).ord
    when 0..127
    when 128
      len = false               # indefinite
    when 255
      fail "reserved length @pos"
    else
      len = be_decode(take(len - 128))
    end
    rtag, rproc = READABLE[[ib, tag]]
    if cons != 0
      nest = []
      if len
        limit = @pos + len
        while @pos < limit
          nest << decode_item
        end
      else
        while (a = decode_item
               if a != [0, 0, ""]
                 nest << a
               end)
          ;
        end
        fail "no indef yet @pos"
      end
      fail [@pos, limit].inspect if @pos != limit
    else
      nest = take(len)
    end
    rproc ? rproc.(nest) : {rtag => nest}
  end

  def decode_item_final
    val = decode_item
    raise "extra bytes follow after a deserialized object" if @pos != @buffer.size
    val
  end

  def decode_item_with_rest
    val = decode_item
    [val, @buffer[@pos..-1]]
  end

  def done
    @pos == buffer.size
  end

  def decode_items
    ret = []
    while @pos != buffer.size
      ret << decode_item
    end
    ret
  end


  end


end
